import random
import typing
import pygame

from . import AgentState
from .navigation import Navigation
from .navigation_agent import NavigationAgent
from .resources import Resources
import shapely as sp
import time


class Robot(NavigationAgent):
    def __init__(self,
                 start_locations: typing.List[typing.Tuple[float, float]],
                 open_locations: typing.List[typing.Tuple[float, float]],
                 navigation: Navigation):
        NavigationAgent.__init__(self,
                                 navigation=navigation,
                                 max_forward_speed=0.075,
                                 max_turning_speed=3.5)
        self.start_locations = start_locations
        self.open_locations = open_locations
        self.last_destination_time = 0

    def reset(self):
        NavigationAgent.reset(self)
        self.set_state(AgentState(location=random.choice(self.start_locations), direction=180))

    @staticmethod
    def create_sprite() -> pygame.Surface:
        sprite = pygame.image.load(Resources.file("predator.png"))
        rotated_sprite = pygame.transform.rotate(sprite, 270)
        return rotated_sprite

    @staticmethod
    def create_polygon() -> sp.Polygon:
        return sp.Polygon([(.02, 0.013), (-.02, 0.013), (-.02, -0.013), (.02, -0.013), (.025, -0.01), (.025, 0.01)])

    def step(self, delta_t: float):
        if self.last_destination_time + 1 < time.time():
            observation = self.get_observation()
            self.last_destination_time = time.time()
            if "prey" in observation["agent_states"] and observation["agent_states"]["prey"]:
                self.set_destination(observation["agent_states"]["prey"][0])

        if not self.path:
            self.set_destination(random.choice(self.open_locations))

        self.navigate(delta_t=delta_t)



import math
import typing

import pygame
from .agent import AgentState, Agent
from .navigation import Navigation
from .navigation_agent import NavigationAgent
from .resources import Resources
import shapely as sp
from enum import Enum
from .util import distance


class MouseObservation(typing.List[float]):
    class Field(Enum):
        prey_x = 0
        prey_y = 1
        prey_direction = 2
        predator_x = 3
        predator_y = 4
        predator_direction = 5
        goal_distance = 6
        predator_distance = 7
        puffed = 8
        puff_cooled_down = 9
        finished = 10

    def __init__(self):
        super().__init__()
        for i in MouseObservation.Field:
            self.append(0.0)

    def __setitem__(self, field: Field, value):
        list.__setitem__(self, field.value, value)


class Mouse(NavigationAgent):
    def __init__(self,
                 start_state: AgentState,
                 actions: typing.List[typing.Tuple[float, float]],
                 goal_location: typing.Tuple[float, float],
                 goal_threshold: float,
                 puff_threshold: float,
                 puff_cool_down_time: float,
                 navigation: Navigation,
                 predator: typing.Optional[Agent] = None):
        NavigationAgent.__init__(self,
                                 navigation=navigation,
                                 max_forward_speed=0.5,
                                 max_turning_speed=20.0)
        self.start_state = start_state
        self.actions = actions
        self.goal_location = goal_location
        self.goal_threshold = goal_threshold
        self.puff_threshold = puff_threshold
        self.puff_cool_down = 0
        self.puff_cool_down_time = puff_cool_down_time
        self.predator = predator
        self.finished = False
        self.puffed = False

    def reset(self):
        self.finished = False
        self.puff_cool_down = 0
        NavigationAgent.reset(self)
        self.set_state(AgentState(location=self.start_state.location,
                                  direction=self.start_state.direction))

    def start(self):
        NavigationAgent.start(self)

    def step(self, delta_t: float):
        if self.puff_cool_down <= 0 and self.predator:
            predator_distance = distance(self.state.location,
                                         self.predator.state.location)
            if predator_distance <= self.puff_threshold:
                if self.model.visibility.line_of_sight(self.state.location, self.predator.state.location):
                    self.puffed = True
                    self.puff_cool_down = self.puff_cool_down_time

        if delta_t < self.puff_cool_down:
            self.puff_cool_down -= delta_t
        else:
            self.puff_cool_down = 0

        goal_distance = distance(self.goal_location, self.state.location)
        self.finished = goal_distance <= self.goal_threshold
        self.navigate(delta_t=delta_t)

    @staticmethod
    def create_sprite() -> pygame.Surface:
        sprite = pygame.image.load(Resources.file("prey.png"))
        rotated_sprite = pygame.transform.rotate(sprite, 270)
        return rotated_sprite

    @staticmethod
    def create_polygon() -> sp.Polygon:
        return sp.Polygon([(.015, 0), (0, 0.005), (-.015, 0), (0, -0.005)])

    def set_action(self, action_number: int):
        self.set_destination(self.actions[action_number])

    def render(self,
               surface: pygame.Surface):
        if self.predator:
            predator_location = self.coordinate_converter.from_canonical(self.model.agents["predator"].state.location)
            puff_area_size = self.model.agents["prey"].puff_threshold * self.coordinate_converter.screen_width
            puff_location = predator_location[0] - puff_area_size, predator_location[1] - puff_area_size
            puff_area_surface = pygame.Surface((puff_area_size * 2, puff_area_size * 2), pygame.SRCALPHA)
            puff_area_color = (255, 0, 0, 60) if self.puff_cool_down > 0 else (0, 0, 255, 60)
            pygame.draw.circle(puff_area_surface,
                               color=puff_area_color,
                               center=(puff_area_size, puff_area_size),
                               radius=puff_area_size)
            surface.blit(puff_area_surface,
                             puff_location)
            pygame.draw.circle(surface=surface,
                               color=(0, 0, 255),
                               center=predator_location,
                               radius=puff_area_size,
                               width=2)
        NavigationAgent.render(self,
                               surface=surface)



from .view import View
from .model import Model
from .agent import Agent, AgentDynamics, AgentState
from .resources import Resources
from .robot import Robot
from .mouse import Mouse, MouseObservation
from .ray_tracing import RayTracing
from .navigation import Navigation
from .util import create_hexagon, distance
from .reward import Reward
from .cellworld_loader import CellWorldLoader



import typing
from .mouse import MouseObservation
class Reward(object):
    def __init__(self,
                 reward_structure: dict):
        self.reward_structure = reward_structure

    def __call__(self, observation: typing.List[float]) -> float:
        reward = 0.0
        for field, multiplier in self.reward_structure.items():
            offset = 0
            if isinstance(multiplier, list):
                offset = multiplier[1]
                multiplier = multiplier[0]
            reward += offset + multiplier * observation[MouseObservation.Field[field].value]
        return reward



import typing

import pygame

from .util import distance, direction, direction_difference, direction_error_normalization
from .agent import Agent
from .navigation import Navigation


class NavigationAgent(Agent):

    def __init__(self,
                 navigation: Navigation,
                 max_forward_speed: float,
                 max_turning_speed: float,
                 threshold: float = 0.01):
        self.max_forward_speed = max_forward_speed
        self.max_turning_speed = max_turning_speed
        self.threshold = threshold
        self.navigation = navigation
        self.destination = None
        self.path = []
        Agent.__init__(self)
        self.collision = False

    def next_step(self):
        if self.path:
            return self.path[0]
        return None

    def set_destination(self, destination):
        if destination != self.destination:
            self.destination = destination
            self.path = self.navigation.get_path(src=self.state.location, dst=self.destination)

    def reset(self):
        self.destination = None
        self.path = []
        Agent.reset(self)

    def start(self):
        Agent.start(self)

    def navigate(self, delta_t: float):

        if self.next_step() is not None:
            distance_error = distance(src=self.state.location,
                                      dst=self.next_step())
            if distance_error < self.threshold:
                self.path.pop(0)

        if self.next_step():
            distance_error = distance(src=self.state.location,
                                      dst=self.next_step())

            normalized_distance_error = max(distance_error/.2, 1)

            destination_direction = direction(src=self.state.location,
                                              dst=self.next_step())
            direction_error = direction_difference(direction1=self.state.direction,
                                                   direction2=destination_direction)
            normalized_direction_error = direction_error_normalization(direction_error=direction_error)

            self.dynamics.forward_speed = self.max_forward_speed * normalized_direction_error * normalized_distance_error
            self.dynamics.turn_speed = self.max_turning_speed * direction_error
        else:
            self.dynamics.forward_speed = 0
            self.dynamics.turn_speed = 0

    def render(self,
               surface: pygame.Surface):
        current_point = self.state.location
        for step in self.path:
            pygame.draw.line(surface,
                             (255, 0, 0),
                             self.coordinate_converter.from_canonical(current_point),
                             self.coordinate_converter.from_canonical(step),
                             2)
            pygame.draw.circle(surface=surface,
                               color=(0, 0, 255),
                               center=self.coordinate_converter.from_canonical(step),
                               radius=5,
                               width=2)
            current_point = step
        Agent.render(self=self, surface=surface)
